% FOIL CALIBRATION PI
% Script for calibration of the resistive touch foil
%% Initialization, open COM ports, open DAQ card
% (Communication has to be closed before re-initializing)

% % set work dir, load previous calibration matrixes
% cd 'C:\Users\zemanji7\Dropbox\Magneticka_platforma\matlab\vision';
% load km4;                                

% open DAQ card
dev = HudaqDevice('MF624');

% open COM port for Magman
% s = serial('COM17');     
% set(s,'BaudRate',230400,'DataBits',8,'Parity','None','StopBits',1);
% fopen(s);
s = mgmopen('COM18'); % Platform
%% Measuring of position for calibration
% Individual coils are consecutively activated to attract the ball. The
% position above each coils is measured and then used for calibration. 
% Evolution of position between coils is recorded as well.

visual = true;                          % visulization durign calibration
numt = 30;                              % number of samples in one step 
nums = 6;                               % number of samples in one step - for calibration of the foil
x = [1 2 3 4 4 3 2 1 1 2 3 4 4 3 2 1];  % sequence of coils for calib
y = [1 1 1 1 2 2 2 2 3 3 3 3 4 4 4 4];
% x = [3 2 3 2];                          % shorter 
% y = [2 2 2 2];
n = numel(x);
pwm = 0.8;                             % pulse width modulation for activated coil

t = zeros(n*numt,1);                    % time
y2 = zeros(n*numt,1);                   % evolution of position in time - resistive foil
x2 = zeros(n*numt,1);
p2 = zeros(n*numt,1);                   % presuure  

xfoil = zeros(n*nums,1);                % positions for calibration - resistive foil
yfoil = zeros(n*nums,1);

mgmoff(s);
tstart = clock;
for cnt = 1:numel(x);
    mgmcoil(s,x(cnt)-1,y(cnt)-1,pwm); 
    for cnt2 = 1:numt;
        tnow = etime(clock, tstart);
        [xf,yf,p] = mgmpos(dev);        
        ind=(cnt-1)*numt+cnt2;
        x2(ind) = xf;
        y2(ind) = yf;
        p2(ind) = p;
        t(ind) = tnow;   
        pause(0.05);
    end
%     for cnt2 = (1:ceil(nums/2))-1 
    for cnt2 = (1:nums)-1
        [xf,yf,p] = mgmpos(dev);                            % read position from the foil
        xfoil(cnt+cnt2*n) = xf; yfoil(cnt+cnt2*n) = yf;
    end
    mgmcoil(s,x(cnt)-1,y(cnt)-1,0);                         % coil turning off
%     for cnt2 = (floor(nums/2):nums)-1 
%         [xf,yf,p] = mgmpos(dev);                            % read position from the foil
%         xfoil(cnt+cnt2*n) = xf; yfoil(cnt+cnt2*n) = yf;
%     end
end
%% Save points for calibration 
save ptscalib x y xfoil yfoil;
%%
plot(t,x2,'LineWidth',2);
grid on;
%% Foil calibration - computation of homography matrix
n = numel(x)*nums;                              % number of points
X = [repmat([x(:),y(:)],[nums 1]),ones(n,1)];   % homogeneous coordinates 
U = [xfoil(:),yfoil(:),ones(n,1)];      % 
C = [U,zeros(n,3),-X(:,1)*ones(1,3).*U;...      % matrix of coorinates
     zeros(n,3),U,-X(:,2)*ones(1,3).*U];
% C=normr(C);                                   % 
[u,sig,v]=svd(C'*C);                              % least square solution of the set of equations
a=u(:,end);                                     % vector of elementd of the homography matrix
km4=reshape(a,[3,3]);                           % homography matrix
xTr=homog(U,km4);                               % transformation of coordinates

% check the transformation
plot(X(:,1),X(:,2),'ro','MarkerSize',10,'LineWidth',2);
axis equal, hold on;
plot(xTr(:,1),xTr(:,2),'bx','MarkerSize',10,'LineWidth',2);
hold off;
grid on;
legend('Reference','Measured')
axis([0,5,0,5]);
%%
save km4 km4
%% Close commuication
fclose(s);
delete(s);
clear s;
Close(dev);
clear all;
