% FOIL CALIBRATION
% Script for calibration of the resistive touch foil
%% Initialization,
rpi = raspi();

%% Measuring of position for calibration
% Individual coils are consecutively activated to attract the ball. The
% position above each coils is measured and then used for calibration. 
% Evolution of position between coils is recorded as well.

visual = true;                          % visulization durign calibration
numt = 30;                              % number of samples in one step 
nums = 6;                               % number of samples in one step - for calibration of the foil
x = [1 2 3 4 4 3 2 1 1 2 3 4 4 3 2 1];  % sequence of coils for calib
y = [1 1 1 1 2 2 2 2 3 3 3 3 4 4 4 4];
% x = [3 2 3 2];                          % shorter 
% y = [2 2 2 2];
n = numel(x);
pwm = 0.9;                             % pulse width modulation for activated coil

t = zeros(n*numt,1);                    % time
y2 = zeros(n*numt,1);                   % evolution of position in time - resistive foil
x2 = zeros(n*numt,1);
p2 = zeros(n*numt,1);                   % presuure  

xfoil = zeros(n*nums,1);                % positions for calibration - resistive foil
yfoil = zeros(n*nums,1);

system(rpi, 'magman/magman.py off')
tstart = clock;
for cnt = 1:numel(x);
    system(rpi, sprintf('magman/magman.py coilxy %d %d %f', x(cnt), y(cnt), pwm))
    fprintf('Ref Pozice %d %d\n', x(cnt), y(cnt))
    for cnt2 = 1:numt;
        tnow = etime(clock, tstart);
        pos = sscanf(system(rpi, 'magman/raspi-ballpos/position.py -p 0'), '%d %d');        
        ind=(cnt-1)*numt+cnt2;
        x2(ind) = pos(1);
        y2(ind) = pos(2);
        p2(ind) = 0;
        fprintf('Pozice %d %d\n', pos)
        t(ind) = tnow;   
        pause(0.05);
    end
    for cnt2 = (1:ceil(nums/2))-1 
        pos = sscanf(system(rpi, 'magman/raspi-ballpos/position.py -p 0'), '%d %d');    % read position from the foil
        xfoil(cnt+cnt2*n) = pos(1); yfoil(cnt+cnt2*n) = pos(2);
    end
    mgmcoil(s,x(cnt)-1,y(cnt)-1,0);                         % coil turning off
    for cnt2 = (floor(nums/2):nums)-1 
        pos = sscanf(system(rpi, 'magman/raspi-ballpos/position.py -p 0'), '%d %d');                            % read position from the foil
        mgmpos3
        xfoil(cnt+cnt2*n) = pos(1); yfoil(cnt+cnt2*n) = pos(2);
    end
end
%% Save points for calibration 
save ptscalib x y xfoil yfoil;
%%
plot(t,x2,'LineWidth',2);
grid on;
%% Foil calibration - computation of homography matrix
n = numel(x)*nums;                              % number of points
X = [repmat([x(:),y(:)],[nums 1]),ones(n,1)];   % homogeneous coordinates 
U = [xfoil(:),yfoil(:),ones(n,1)];              % 
C = [U,zeros(n,3),-X(:,1)*ones(1,3).*U;...      % matrix of coorinates
     zeros(n,3),U,-X(:,2)*ones(1,3).*U];
% C=normr(C);                                   % 
[u,sig,v]=svd(C'*C);                              % least square solution of the set of equations
a=u(:,end);                                     % vector of elementd of the homography matrix
km4pi=reshape(a,[3,3]);                          % homography matrix
xTr=homog(U,km4pi);                              % transformation of coordinates
% check the transformation
plot(X(:,1),X(:,2),'ro','MarkerSize',10,'LineWidth',2);
axis equal, hold on;
plot(xTr(:,1),xTr(:,2),'bx','MarkerSize',10,'LineWidth',2);
hold off;
grid on;
legend('Reference','Measured')
axis([0,5,0,5]);
%%
save km4pi km4pi
%% Close commuication
fclose(s);
delete(s);
clear s;

clear all;
