#if defined(MATLAB_MEX_FILE)
#include "tmwtypes.h"
#include "simstruc_types.h"
#else
#include "rtwtypes.h"
#endif

/* %%%-SFUNWIZ_wrapper_includes_Changes_BEGIN --- EDIT HERE TO _END */
# ifndef MATLAB_MEX_FILE

#include <wiringPi.h>
#include <wiringSerial.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <termios.h>

int fp = -1;

int magman_open(char *portname){
    if(fp==-1) fp = serialOpen(portname, 230400);
    
    struct termios options ;

    tcgetattr (fp, &options) ;   // Read current options
    options.c_lflag &= ~ICANON; /* Set non-canonical mode */
    options.c_cc[VTIME] = 1; /* Set timeout of 10.0 seconds */
    tcsetattr (fp, TCSANOW, &options) ;   // Set new options
    
    return fp;
}

void magman_sendPacket(unsigned char *packet, int len){
    write(fp, packet, len);
    //serialFlush(fp); 
}

unsigned short simpleCrc(char *buff, unsigned count)
{
  unsigned short work = 0;
  unsigned char a, j, k;
  for (k = 0; k < count; k++) {
    a = buff[k];
    work ^= a;
    for (j = 0; j < 8; ++j) {
      if (work & 1)
        work = (work >> 1) ^ 0xA001;
      else
        work = (work >> 1);
    }
  }
  return work;
}

int magman_setModulePacket(int module, real_T input[], unsigned char *packet){
    int crc_val=0;
    int i;

    real_T channel[4];

    for(i=0;i<4;i++){
        channel[i] = input[i];
        if(channel[i] >  1) channel[i] = 1;
        if(channel[i] < -1) channel[i] = -1;
        if(channel[i] >= 0){
            channel[i] = channel[i]*2047;
        }else{
            channel[i] = 0xFFF+channel[i]*2047;
        }
        //channel[i] = channel[i] & 0xFFF;
    }

    packet[0] = 1;
    packet[1] = 0x80 | module;
    packet[2] = 0x91;

    sprintf(packet+3, "%03X%03X%03X%03X\n", (int)channel[0], (int)channel[1], (int)channel[2], (int)channel[3]);

    packet[15] = 0x04;

    crc_val = simpleCrc(packet, 16);

    packet[16] = crc_val & 0xFF;
    packet[17] = crc_val >> 8;
    return 18;
}

int magman_getPosition(real_T *x, real_T *y, real_T *press){
    char req[] = {0x01, 0xbf, 0xb9, 0x04, 0x42, 0x4b};
    char rep[17];
    char i;
    int ret;
    ret = write(fp, req, 6);
    if(ret!=6){
        return -1;
    }
    ret = read(fp, rep, 16);
    if(ret!=16){
        return -2;
    }
    rep[17]=0;
    
    int rcrc = rep[14] | rep[15]<<8;
    int crc = simpleCrc(rep, 14);

    if(crc!=rcrc){
        *x=0;
        *y=0;
        *press=0;
        return -3;
    }

    rep[13]=0;
    *press = strtol(rep+9, NULL, 16);
    rep[9]=0;
    *y = strtol(rep+6, NULL, 16)/4096.0;
    rep[6]=0;
    *x = strtol(rep+3, NULL, 16)/4096.0;
    serialFlush(fp);
    return 0;
}

int magman_getPositionCheck(real_T *x, real_T *y, real_T *press){
    char req[] = {0x01, 0xbf, 0xb9, 0x04, 0x42, 0x4b};
    char rep[17];
    char i;    
    int ret;
    int bytes;
    ioctl(fp, FIONREAD, &bytes);
    read(fp, rep, bytes);
    
    ret = write(fp, req, 6);
    if(ret!=6){
        //return -1;
    }
    
    if(read(fp, rep, 1)!=1){
        //return -2;
    }
    if(rep[0] != 2){
        //return -3;
    }
    ret=1;
    for(i=0; i<15; i++){
        if(read(fp, rep+i+1, 1)!=1){
            return ret;
        }
        ret++;
    }
    rep[16]=0;
    
    int rcrc = rep[14] | rep[15]<<8;
    int crc = simpleCrc(rep, 14);

    if(crc!=rcrc){
        *x=0;
        *y=0;
        *press=0;
        return ret;
    }

    rep[13]=0;
    *press = strtol(rep+9, NULL, 16);
    rep[9]=0;
    *y = strtol(rep+6, NULL, 16)/4096.0;
    rep[6]=0;
    *x = strtol(rep+3, NULL, 16)/4096.0;
    return ret;
}
#endif